# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validate an analysis template."""

import json
import logging
from typing import Dict, List

from Bio.SeqRecord import SeqRecord
from goodtables import validate
from importlib_resources import open_text
from pandas import DataFrame

from . import schemata


__all__ = ("validate_template", "drop_missing_records")

LOGGER = logging.getLogger(__name__)
with open_text(schemata, "template.json", encoding="utf-8") as file_handle:
    TEMPLATE_SCHEMA = json.load(file_handle)


def validate_template(template: DataFrame) -> List[Dict]:
    """
    Validate that the template respects the desired schema.

    Make use of goodtables [1]_ in order to validate the template data frame
    against the chosen specification.

    Parameters
    ----------
    template : pandas.DataFrame
        A data frame with three columns: 'plasmid', 'primer', 'sample'.

    Returns
    -------
    list
        List of errors that are themselves dictionaries with the keys 'code'
        and 'message'.

    References
    ----------
    .. [1] https://pypi.org/project/goodtables/

    """
    if len(template) == 0:
        return [
            {
                "code": "empty",
                "message": "There must be at least one data row."
            }
        ]
    # Convert the data frame into a format suitable for goodtables, the keys
    # of the first record are taken as the column headers.
    records = template.to_dict("records")
    result = validate(
        records, preset="table", headers=list(records[0]),
        schema=TEMPLATE_SCHEMA, order_fields=True)["tables"][0]
    return result.get("errors", [])


def drop_missing_records(
    template: DataFrame,
    plasmids: Dict[str, SeqRecord],
    samples: Dict[str, SeqRecord]
) -> DataFrame:
    """
    Drop rows with missing sequence records from the template.

    Template rows that lack a corresponding plasmid sequence or sample
    sequence are removed from the template thus ensuring a consistent analysis.

    Parameters
    ----------
    template : pandas.DataFrame
        A data frame with three columns: 'plasmid', 'primer', 'sample'.
    plasmids : dict
        A dictionary mapping plasmid names to their sequence records.
    samples : dict
        A dictionary mapping sample names to their sequence records.

    Returns
    -------
    pandas.DataFrame
        Return the template only for those rows where sequences exist.

    """
    plasmid_mask = template["plasmid"].isin(plasmids)
    if (~plasmid_mask).any():
        LOGGER.error(
            "The following plasmid(s) have no corresponding sequence record: "
            "%s.", ", ".join(template.loc[plasmid_mask, "plasmid"]))
    sample_mask = template["sample"].isin(samples)
    if (~sample_mask).any():
        LOGGER.error(
            "The following sample(s) have no corresponding sequence record: "
            "%s.", ", ".join(template.loc[sample_mask, "sample"]))
    return template.loc[plasmid_mask & sample_mask, :]



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validate a plasmid sequence record."""

import logging
from typing import Iterable

from Bio.SeqRecord import SeqRecord


__all__ = ("validate_plasmid",)

LOGGER = logging.getLogger(__name__)


def validate_plasmid(plasmid: SeqRecord, primer_ids: Iterable[str]):
    """
    Validate that the plasmid sequence record has the required annotations.

    Parameters
    ----------
    plasmid : Bio.SeqRecord.SeqRecord
        A plasmid sequence record that must contain the following annotations.
    primer_ids : iterable
        The given primer identifiers coming from the analysis template should
        be annotated in the plasmid sequence record.

    Returns
    -------
    list
        List of errors that are themselves dictionaries with the keys 'code'
        and 'message'.

    """
    assert len(plasmid) > 0
    assert hasattr(plasmid, "features")
    for feat in plasmid.features:
        assert hasattr(feat, "location")
        assert hasattr(feat.location, "start")
        assert hasattr(feat.location.start, "position")
        assert 0 <= feat.location.start.position <= len(plasmid)
        assert hasattr(feat.location, "end")
        assert hasattr(feat.location.end, "position")
        assert 0 <= feat.location.end.position <= len(plasmid)
        assert hasattr(feat, "type")
        assert hasattr(feat, "qualifiers")
        assert "label" in feat.qualifiers
    for primer_id in primer_ids:
        # TODO: Test that all given primer identifiers are on the plasmid.
        pass



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide sample analysis functions."""

import logging

from Bio.SeqRecord import SeqRecord
from numpy import arange, array, asarray, nanmedian

from ..config import Configuration


__all__ = ("trim_sample",)

LOGGER = logging.getLogger(__name__)


def trim_sample(seq: SeqRecord) -> (int, SeqRecord, array, int, float):
    """Cut off low quality ends of a Sanger sequencing record."""
    LOGGER.debug("Trim sample.")
    config = Configuration()
    scores = asarray(seq.letter_annotations['phred_quality'])
    median = float(nanmedian(scores))
    if median < config.threshold:
        message = (
            f"The median Phred quality ({median}) is below the "
            f"required threshold ({config.threshold}).")
        LOGGER.error(message)
        raise ValueError(message)
    mask = (scores >= config.threshold)
    index = arange(len(mask), dtype=int)
    min_i = index[mask][0]
    max_i = index[mask][-1] + 1  # Since Python excludes upper range limit.
    LOGGER.debug("Cutting %d nucleotides at the beginning and %d at the end.",
                 min_i + 1, len(seq) - max_i)
    return (
        min_i, seq[min_i:max_i], scores[min_i:max_i], len(seq) - max_i, median)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A basic client for the jbei ICE API."""

from .repository_client import RepositoryClient


__all__ = ("ICEClient",)


class ICEClient(RepositoryClient):
    """Provide a pythonic interface to the ICE API."""

    def __init__(self):
        """Initialize the ICE client."""
        super().__init__()

    def get_plasmid_ids(self):
        """Return a frozenset of all accessible plasmid identifiers."""
        raise NotImplementedError("Coming soon.")

    def get_plasmid_record(self, plasmid_id):
        """Return the name and sequence record for a specific plasmid."""
        raise NotImplementedError("Coming soon.")

    def get_primer_ids(self):
        """Return a frozenset of all accessible primer identifiers."""
        raise NotImplementedError("Coming soon.")



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validate a sample sequence record."""

import logging

from Bio.SeqRecord import SeqRecord


__all__ = ("validate_sample",)

LOGGER = logging.getLogger(__name__)


def validate_sample(sample: SeqRecord):
    """
    Validate that the sample sequence record has Phred quality scores.

    Parameters
    ----------
    sample : Bio.SeqRecord.SeqRecord
        A sample sequence record that must contain the following annotations:

    Returns
    -------
    list
        List of errors that are themselves dictionaries with the keys 'code'
        and 'message'.

    """
    assert hasattr(sample, "letter_annotations")
    assert "phred_quality" in sample.letter_annotations
    assert len(sample.letter_annotations["phred_quality"]) == len(sample)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide Sanger sequence alignment functions and scoring."""

import logging
import re
from os.path import join
from typing import List, get_type_hints

from Bio import AlignIO
from Bio.Emboss.Applications import WaterCommandline
from Bio.SeqRecord import SeqRecord
from numpy import array, nan
from pandas import DataFrame

from ..config import Configuration


__all__ = ("emboss_alignment", "alignment_to_table")

LOGGER = logging.getLogger(__name__)


def alignment_to_table(align: AlignIO.MultipleSeqAlignment,
                       scores: array,
                       start: int) -> DataFrame:
    """Generate a table of the alignment."""
    report = list()
    b_coord = align.positions['bseq_start']
    a_coord = align.positions['aseq_start']
    for a_char, b_char in zip(align[0], align[1]):
        if a_char == b_char:
            report.append((a_coord, b_coord, False, a_char, b_char))
            a_coord += 1
            b_coord += 1
        elif b_char == '-':
            report.append((a_coord, nan, True, a_char, b_char))
            a_coord += 1
        elif a_char == '-':
            report.append((nan, b_coord, True, a_char, b_char))
            b_coord += 1
        else:
            report.append((a_coord, b_coord, True, a_char, b_char))
            a_coord += 1
            b_coord += 1
    df = DataFrame(report, columns=[
        'plasmid_pos', 'sample_pos', 'snp', 'plasmid_chr', 'sample_chr'])
    # Subtract 1 for Python indexing into scores.
    df["quality"] = (
        df.loc[df["sample_pos"].notnull(), "sample_pos"].astype(int) - 1).map(
        scores.__getitem__)
    # Re-align to original sample sequence position.
    df.loc[df["sample_pos"].notnull(), "sample_pos"] += start
    return df


def extract_emboss_positions(lines: List[str]):
    """Extract the alignment positions from the `water` output file."""
    asis_lines = [l.strip() for l in lines if l.startswith('asis')]
    start_pattern = re.compile(r'^asis\s+(\d+)')
    end_pattern = re.compile(r'(\d+)$')
    return {
        'aseq_start': int(start_pattern.match(asis_lines[0])[1]),
        'bseq_start': int(start_pattern.match(asis_lines[1])[1]),
        'aseq_end': int(end_pattern.search(asis_lines[-2])[1]),
        'bseq_end': int(end_pattern.search(asis_lines[-1])[1])
    }


def emboss_alignment(sample_id: str,
                     sample_sequence: SeqRecord,
                     plasmid_id: str,
                     plasmid_sequence: SeqRecord,
                     gap_open_penalty: float=2.0,
                     gap_extension_penalty: float=10.0,
                     tool: get_type_hints(WaterCommandline)=WaterCommandline
                     ) -> AlignIO.MultipleSeqAlignment:
    """
    Create an alignment between the known plasmid sequence and the Sanger read.

    We generally expect the sequences to be identical excluding single
    nucleotide errors coming from real mutations, replication errors, or bad
    reads. Read quality is not yet taken into account but will in future.

    Parameters
    ----------
    plasmid_sequence
    sample_sequence
    sample_id
    plasmid_id
    gap_open_penalty
    gap_extension_penalty
    tool

    Returns
    -------
    Bio.AlignIO.MultipleSeqAlignment
        The pairwise alignment.

    """
    config = Configuration()
    # Possibly the output of `aformat="markx10"` is easier to parse.
    cmd = tool(gapopen=gap_open_penalty, gapextend=gap_extension_penalty)
    cmd.asequence = f"asis:{plasmid_sequence.seq}"
    cmd.bsequence = f"asis:{sample_sequence.seq}"
    outfile = join(config.output, f"{sample_id}_{plasmid_id}.txt")
    cmd.outfile = outfile
    stdout, stderr = cmd()
    LOGGER.debug(stdout)
    LOGGER.debug(stderr)  # Unfortunately, normal messages are sent to stderr...

    align = AlignIO.read(cmd.outfile, "emboss")
    # Get coordinates manually because biopython ...
    identity = align.annotations['identity'] / len(sample_sequence)
    LOGGER.debug("Sequence identity is %0.2g.", identity)
    if identity < 0.9:
        LOGGER.info("Trying reverse complement!")
        cmd.bsequence = f"asis:{sample_sequence.reverse_complement().seq}"
        rev_outfile = join(
            config.output, "{sample_id}_{plasmid_id}_rev.txt")
        cmd.outfile = rev_outfile
        stdout, stderr = cmd()
        LOGGER.debug(stdout)
        LOGGER.debug(stderr)
        rev_align = AlignIO.read(cmd.outfile, "emboss")
        rev_identity = rev_align.annotations['identity'] / len(sample_sequence)
        LOGGER.debug("Complement sequence identity is %0.2g.", rev_identity)
    else:
        rev_identity = -1.0
        rev_align = None
        rev_outfile = None
    if identity > rev_identity:
        with open(outfile) as file_h:
            align.positions = extract_emboss_positions(file_h.readlines())
        LOGGER.debug(str(align.positions))
        return align
    else:
        with open(rev_outfile) as file_h:
            rev_align.positions = extract_emboss_positions(file_h.readlines())
        LOGGER.debug(str(rev_align.positions))
        return rev_align



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide a high-level interface for the Sanger sequence analysis workflow.

Functionality to retrieve identifiers and sequence records from repositories
can be imported from the ``sanger_sequencing.clients`` subpackage.
"""

import logging
from typing import Dict, List

from Bio.SeqRecord import SeqRecord
from pandas import DataFrame

from . import analysis, validation
from .helpers import log_errors


__all__ = ("sanger_verification", "plasmid_report", "sample_report")

LOGGER = logging.getLogger(__name__)


def sanger_verification(template: DataFrame,
                        plasmids: Dict[str, SeqRecord],
                        samples: Dict[str, SeqRecord]
                        ) -> List[Dict]:
    """
    Perform a complete Sanger verification for many plasmids and sample reads.

    Parameters
    ----------
    template : pandas.DataFrame
        A template table with three columns: plasmid, primer, sample which
        are all identifiers.
    plasmids : dict
        A mapping from plasmid identifiers to sequence records.
    samples : dict
        A mapping from sample identifiers to sequence records.

    Returns
    -------
    list
        A list of dictionaries that are each a plasmid report.

    Raises
    ------
    AssertionError
        All function arguments are extensively validated and may raise errors.

    See Also
    --------
    plasmid_report

    """
    LOGGER.info("Validate template.")
    errors = validation.validate_template(template)
    if len(errors) > 0:
        log_errors(errors)
        raise AssertionError("Invalid analysis template.")
    LOGGER.info("Validate plasmids.")
    for plasmid in plasmids.values():
        validation.validate_plasmid(plasmid, [])
    LOGGER.info("Validate samples.")
    for sample in samples.values():
        validation.validate_sample(sample)
    template = validation.drop_missing_records(template, plasmids, samples)
    LOGGER.info("Generate reports.")
    return [plasmid_report(plasmid_id, plasmids[plasmid_id], sub, samples)
            for plasmid_id, sub in template.groupby(
            "plasmid", as_index=False, sort=False)]


def plasmid_report(plasmid_id: str,
                   sequence: SeqRecord,
                   template: DataFrame,
                   samples: Dict[str, SeqRecord]) -> Dict:
    """
    Create an analysis report for a single plasmid and one or more reads.

    The plasmid report contains detailed reports on each sample read
    performed. It then evaluates each sequence alignment conflict using
    information from other samples as appropriate.

    Parameters
    ----------
    plasmid_id : str
        The plasmid identifier.
    sequence : Bio.SeqRecord.SeqRecord
        The plasmid's sequence record.
    template : pandas.DataFrame
        A part of the template table concerning this plasmid only.
    samples : dict
        A mapping from sample identifiers to sequence records.

    Returns
    -------
    dict
        An individual plasmid report.

    """
    LOGGER.info("Analyze plasmid '%s'.", plasmid_id)
    report = {
        "id": plasmid_id,
        "name": sequence.name,
        "samples": [
            sample_report(row.sample, samples[row.sample], row.primer,
                          plasmid_id, sequence)
            for row in template.itertuples(index=False)]
    }
    # Post-process reports in order to classify conflicts.
    LOGGER.debug("Concatenate the detailed sample reports.")
    total = analysis.concatenate_sample_reports(report["samples"])
    for rep in report["samples"]:
        rep["conflicts"] = analysis.summarize_plasmid_conflicts(
            rep["details"], total, sequence)
    return report


def sample_report(sample_id: str,
                  sample_sequence: SeqRecord,
                  primer_id: str,
                  plasmid_id: str,
                  plasmid_sequence: SeqRecord) -> Dict:
    """
    Create an analysis report for a single sample read.

    Parameters
    ----------
    sample_id : str
        The sample identifier.
    sample_sequence :  Bio.SeqRecord.SeqRecord
        The sample's sequence record.
    primer_id : str
        The primer identifier.
    plasmid_id : str
        The plasmid identifier.
    plasmid_sequence : Bio.SeqRecord.SeqRecord
        The plasmid's sequence record.

    Returns
    -------
    dict
        An individual sample report.

    """
    LOGGER.info("Analyze sample '%s'.", sample_id)
    report = {
        "id": sample_id,
        "primer": primer_id,
        "readLength": len(sample_sequence),
        "errors": []
    }
    # Convert to base `float` for JSON compatibility.
    try:
        start, trimmed_seq, quality_scores, end, median = analysis.trim_sample(
            sample_sequence)
    except ValueError as err:
        report["errors"].append(str(err))
        return report
    report["medianQuality"] = median
    report["cutBeginning"] = int(start)
    report["cutEnd"] = int(end)
    align = analysis.emboss_alignment(
        sample_id, trimmed_seq, plasmid_id, plasmid_sequence)
    report["details"] = analysis.alignment_to_table(
        align, quality_scores, start)
    return report



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide a global configuration singleton."""

import logging
from tempfile import mkdtemp


__all__ = ("Configuration",)

LOGGER = logging.getLogger(__name__)


class Singleton(type):
    """Implementation of the singleton pattern."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs)
        return cls._instances[cls]


class Configuration(metaclass=Singleton):
    """
    Configure values that affect how the analysis is performed.

    Attributes
    ----------
    threshold : float
        Threshold on the Phred quality. The Phred score scales between 0 and
        62. A typical good Sanger sequencing read has a score of 55.
    output : str or pathlib.Path
        Output directory for alignment files.

    """

    def __init__(self, threshold=50.0, output=mkdtemp(), **kwargs):
        """
        Initialize the singleton configuration object.

        Parameters
        ----------
        threshold : float, optional
            Threshold on the Phred quality. The Phred score scales between 0 and
            62. A typical good Sanger sequencing read has a score of 55.
        output : str or pathlib.Path, optional
            Output directory for alignment files.
        kwargs : dict
            Passed to the ``super()`` class.

        """
        super().__init__(**kwargs)
        self.threshold = threshold
        self.output = output



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions that drive the main Sanger sequence analysis workflow."""

import logging
from typing import Dict, List

from Bio.Data.CodonTable import ambiguous_dna_by_name
from Bio.SeqRecord import SeqRecord
from numpy import isnan, nanmean
from pandas import DataFrame, concat

from ..config import Configuration


__all__ = ("summarize_plasmid_conflicts", "concatenate_sample_reports")

LOGGER = logging.getLogger(__name__)
CODON_TABLE = ambiguous_dna_by_name["Standard"].forward_table
START_CODONS = frozenset(ambiguous_dna_by_name["Standard"].start_codons)
STOP_CODONS = frozenset(ambiguous_dna_by_name["Standard"].stop_codons)


def concatenate_sample_reports(reports):
    """Concatenate many data frames into one."""
    data = []
    for sample in reports:
        df = sample.get("details")
        if df is None:
            continue
        df["sample"] = sample["id"]
        df["primer"] = sample["primer"]
        data.append(df)
    return concat(data, ignore_index=True, copy=False)


def determine_type(row):
    if isnan(row.plasmid_pos) and isnan(row.sample_pos):
        msg = ("Detected a gap in the alignment. Only expecting single "
               "position conflicts.")
        LOGGER.error(msg)
        raise ValueError(msg)
    elif isnan(row.plasmid_pos):
        return "insertion"
    elif isnan(row.sample_pos):
        return "deletion"
    else:
        return "change"


def determine_quality(region, threshold):
    if nanmean(region["quality"]) >= threshold:
        return "high"
    else:
        return "low"


def confirm_conflict(conflict_type, row, cover, threshold):
    num_confirmed = 0
    num_invalidated = 0
    for sample_id, sub in cover.groupby("sample", as_index=False,
                                        sort=False):
        if determine_quality(sub, threshold) == "low":
            LOGGER.debug(
                "Ignoring low quality sample region for conflict "
                "confirmation.")
            continue
        # Index 1 should correspond to the mid-point and thus to the
        # conflict site. However, due to sequence read trimming we may hit
        # the end of the sequence alignment.
        if len(sub) < 3:
            LOGGER.debug(
                "Ignoring incomplete sample region (less than three "
                "positions).")
            continue
        cmp = sub.iloc[1]
        if not cmp.snp:
            LOGGER.debug("Not a conflict site.")
            num_invalidated += 1
            continue
        if determine_type(cmp) != conflict_type:
            LOGGER.debug("Different type of conflict site.")
            num_invalidated += 1
            continue
        if (cmp.sample_chr == row.sample_chr) and \
                (cmp.plasmid_chr == row.plasmid_chr):
            num_confirmed += 1
        else:
            num_invalidated += 1
    return num_confirmed, num_invalidated


def determine_effects(row, plasmid, previous, following):
    """
    Post-process conflicts and categorize them.

    Parameters
    ----------
    row
    plasmid
    previous
    following

    Returns
    -------
    list, list

    """
    features = []
    effects = []
    for feat in plasmid.features:
        # Process only features if the conflict position overlaps.
        if not (previous <= feat.location.end.position and
                feat.location.start.position <= following):
            continue
        features.append((feat.type, feat.qualifiers["label"]))
        if feat.type == "CDS":
            # Potential frame shift (usually rather a sequencing error).
            if isnan(row.plasmid_pos) or isnan(row.sample_pos):
                effects.append("Frame shift")
                continue
            # Does the new codon cause an amino acid change?
            seq = feat.extract(plasmid)
            # Translate plasmid position into feature position.
            feat_pos = int(row.plasmid_pos) - feat.location.start.position
            feat_pos -= 1  # Transform to 0-indexing.
            codon_pos = feat_pos % 3
            codon = list(seq[
                         (feat_pos - codon_pos):(
                             feat_pos - codon_pos + 3)])
            if len(codon) < 3:
                LOGGER.error(
                    "SNP at the beginning or end of CDS. "
                    "Unknown effect.")
                effects.append("Unknown")
                continue
            cdn = "".join(codon)
            try:
                plasmid_aa = CODON_TABLE[cdn]
            except (KeyError, ValueError):
                if cdn in START_CODONS:
                    LOGGER.warning("Start codon hit on plasmid.")
                    plasmid_aa = "START"
                elif cdn in STOP_CODONS:
                    LOGGER.warning("Stop codon hit on plasmid.")
                    plasmid_aa = "STOP"
                else:
                    LOGGER.error("Unknown codon '%s' on plasmid.", cdn)
                    plasmid_aa = "UNKNOWN"
            codon[codon_pos] = row.sample_chr
            cdn = "".join(codon)
            try:
                sample_aa = CODON_TABLE[cdn]
            except (KeyError, ValueError):
                if cdn in START_CODONS:
                    LOGGER.warning("Change to start codon on sample.")
                    sample_aa = "START"
                elif cdn in STOP_CODONS:
                    LOGGER.warning("Change to stop codon on sample.")
                    sample_aa = "STOP"
                else:
                    LOGGER.error("Unknown codon '%s' on sample.", cdn)
                    sample_aa = "UNKNOWN"
            effects.append(f"{plasmid_aa} -> {sample_aa}")
    return features, effects


def summarize_plasmid_conflicts(sample: DataFrame,
                                total: DataFrame,
                                plasmid: SeqRecord) -> List[Dict]:
    """
    Add useful information on sequence conflicts and their surroundings.

    There are three principle types of conflicts compared to a reference
    sequence:
    1. insertions
    2. deletions
    3. changes

    Depending on the quality in the neighborhood and information from other
    samples, each conflict may be (highly) likely, unresolved (i.e.,
    an unclear situation), or resolved (invalidated by other samples).

    """
    config = Configuration()
    conflicts = []
    # Show what happens around a mismatch location on all samples.
    LOGGER.info("Assessing %d conflicts.", total["snp"].sum())
    for row in sample.loc[sample["snp"], :].itertuples():
        conflict = row._asdict()
        del conflict["snp"]
        del conflict["Index"]
        # The assumption here is that there will ever only be single position
        # conflicts and we can check the sequence before and after for more
        # information. This probably holds for a good read. Bad reads should
        # be repeated and not analyzed automatically.
        region = total.iloc[[row.Index - 1, row.Index, row.Index + 1], :]
        # Determine the kind of conflict.
        try:
            conflict["type"] = determine_type(row)
        except ValueError:
            continue
        # Determine the quality of the region.
        conflict["surroundingQuality"] = determine_quality(
            region, config.threshold)
        # Check for more information on other samples.
        # Due to a potential gap we take the plasmid index position before and
        # check rows in the total in-between that index and index + 2 which
        # should cover the gap.
        index = total[
            (total["sample"] != row.sample) &
            (total["plasmid_pos"] == region["plasmid_pos"].min())].index
        index = [j for i in index for j in range(i, i + 3)]
        cover = total.loc[index, :]
        conflict["confirmed"], conflict["invalidated"] = confirm_conflict(
            conflict["type"], row, cover, config.threshold)
        # Add feature data.
        conflict["featuresHit"], conflict["effect"] = determine_effects(
            row, plasmid, region["plasmid_pos"].min(),
            region["plasmid_pos"].max())
        conflicts.append(conflict)
    return conflicts



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A basic client for the LabCollector API v2."""

import base64
from functools import lru_cache
from io import StringIO
from operator import itemgetter
from typing import FrozenSet, Tuple
from urllib.parse import urljoin

import requests
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from .repository_client import RepositoryClient


__all__ = ("LabCollectorClient",)


class LabCollectorClient(RepositoryClient):
    """Provide a pythonic interface to the LabCollector API v2."""

    COUNT_PARAMETERS = {
        "fields": "count"
    }
    FIELD_PARAMETERS = {
        "fields": "Sequence_file_GenBank"
    }

    def __init__(self,
                 api: str,
                 token: str,
                 timeout: float=10.0,
                 cache_size: int=10_000,
                 **kwargs):
        """
        Initialize a Labcollector API v2 client.

        Parameters
        ----------
        api : str
            The base Labcollector API URL; typically similar to
            https://labcollector.your.domain/webservice/v2/.
        token : str
            A token that has read access to plasmids and primers.
        timeout : float
            A timeout in seconds for the GET requests.
        cache_size : int, optional
            The maximal size of the LRU cache. A trade-off between
            performance for already requested resources and the memory used.
        kwargs : optional
            Additional keyword arguments are used to update the default
            header for making requests to the API.

        """
        super().__init__()
        self.api = api
        if not self.api.endswith("/"):
            self.api += "/"
        self.token = token
        self.timeout = timeout
        self.headers = {
            'X-LC-APP-Auth': token,
            'Accept': "application/json",
            'Cache-Control': "no-cache"
        }
        self.headers.update(kwargs)
        self.plasmids_resource = urljoin(self.api, "plasmids")
        self.primers_resource = urljoin(self.api, "primers")
        # We want the instance methods to be cached rather than the class
        # methods.
        self.get_plasmid_ids = lru_cache(maxsize=1)(self.get_plasmid_ids)
        self.get_plasmid_record = lru_cache(maxsize=int(cache_size))(
            self.get_plasmid_record)
        self.get_primer_ids = lru_cache(maxsize=1)(self.get_primer_ids)

    def _get_resource_ids(self, endpoint: str) -> FrozenSet:
        response = requests.get(
            endpoint, headers=self.headers, params=self.COUNT_PARAMETERS,
            timeout=self.timeout)
        response.raise_for_status()
        # The counter seems to be the sequential identifier for objects
        # within the Labcollector database.
        count_getter = itemgetter("count")
        return frozenset(count_getter(elem) for elem in response.json())

    def get_plasmid_ids(self) -> FrozenSet:
        """Return a frozenset of all accessible plasmid identifiers."""
        return self._get_resource_ids(self.plasmids_resource)

    def get_primer_ids(self) -> FrozenSet:
        """Return a frozenset of all accessible primer identifiers."""
        return self._get_resource_ids(self.primers_resource)

    def _get_sequence_record(self,
                             endpoint: str,
                             record_id: str) -> Tuple[str, SeqRecord]:
        endpoint += f"/{record_id}"
        response = requests.get(
            endpoint, headers=self.headers, params=self.FIELD_PARAMETERS,
            timeout=self.timeout)
        response.raise_for_status()
        data = response.json()[0]
        name = data["Sequence_file_GenBank"][0]["name"]
        content = data["Sequence_file_GenBank"][0]["content"]
        return name, SeqIO.read(
            StringIO(base64.b64decode(content).decode("ascii")), "gb")

    def get_plasmid_record(self, plasmid_id: str) -> Tuple[str, SeqRecord]:
        """
        Return the name and sequence record for a specific plasmid.

        Parameters
        ----------
        plasmid_id : str
            The plasmid identifier of interest.

        Returns
        -------
        (str, Bio.SeqRecord.SeqRecord)
            The plasmid name and sequence record if any.

        """
        return self._get_sequence_record(self.plasmids_resource, plasmid_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability, Technical University Denmark
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup the sanger_sequencing package."""

import versioneer
from setuptools import setup


# All other arguments are defined in `setup.cfg`.
setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    # Temporary workaround for https://github.com/pypa/setuptools/issues/1136.
    package_dir={"": "src"}
)
